import json
import time
import logging
from functools import wraps
from math import trunc

from connexity.CONST import CONNEXITY_METRICS_URL
from connexity.utils.send_data import send_data
from fastapi.responses import StreamingResponse
from connexity.metrics.schemas import ElevenlabsRequestBody
from connexity.metrics.utils import contains_full_sentence

logger = logging.getLogger(__name__)


def measure_first_chunk_latency():
    """
    Decorator for FastAPI endpoints that return a StreamingResponse.
    Logs the time to the first sentence (TTFS) from when the endpoint
    function is invoked.
    """
    def actual_decorator(method):
        @wraps(method)
        async def wrapper(*args, **kwargs):
            start_time = time.time()  # Mark start of endpoint function
            data: ElevenlabsRequestBody = kwargs.get("request_body")
            if data:
                print(data.elevenlabs_extra_body, flush=True)
                sid = data.elevenlabs_extra_body.get("sid")
            else:
                sid = None
            original_response = await method(*args, **kwargs)

            # If the response is not StreamingResponse, just return it
            if not isinstance(original_response, StreamingResponse):
                return original_response

            # Wrap the original response's body_iterator
            async def wrapped_body_iterator():
                first_sentence = True
                sentence = ""
                async for chunk in original_response.body_iterator:
                    try:
                        if first_sentence:
                            data = chunk[5:]
                            data = json.loads(data)
                            sentence += data.get("choices")[0].get("delta").get("content")
                            if contains_full_sentence(sentence):
                                latency = time.time() - start_time
                                print(sentence, flush=True)
                                print(f"[measure_first_chunk_latency] Time to first chunk: {latency:.4f} seconds", flush=True)
                                first_sentence = False
                                data_dict = {"sid": sid,
                                             "latency": trunc(latency * 1000),
                                             "first_sentence": sentence}
                                await send_data(data_dict, api_key='none', url=CONNEXITY_METRICS_URL)
                    except Exception as e:
                        print(f"Exception has occur while processing chunks {e.args}", flush=True)
                    yield chunk

            # Create a new StreamingResponse that uses our wrapped iterator
            new_response = StreamingResponse(
                wrapped_body_iterator(),
                media_type=original_response.media_type
            )

            # Preserve status code and headers from the original response
            new_response.status_code = original_response.status_code
            new_response.headers.update(original_response.headers)

            return new_response

        return wrapper

    return actual_decorator



from typing import Optional

from connexity.calls.messages.interface_message import InterfaceMessage


class ToolCallMessage(InterfaceMessage):

    def __init__(self, content: str,
                 tool_call_id: str,
                 name: str,
                 seconds_from_start: Optional[float] = None,
                 arguments: Optional[dict] = None
                 ):
        super().__init__("tool_call", content, seconds_from_start)
        self.tool_call_id = tool_call_id
        self.name = name
        self.arguments = arguments

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "role": self.role,
            "content": self.content,
            "seconds_from_start": self.seconds_from_start,
            "tool_call_id": self.tool_call_id,
            "name": self.name,
            "arguments": self.arguments
        }



from typing import Optional

from connexity.calls.messages.interface_message import InterfaceMessage


class ToolResultMessage(InterfaceMessage):

    def __init__(self, content: str,
                 tool_call_id: str,
                 result_type: str,
                 seconds_from_start: Optional[float] = None,
                 result: Optional[str] = None
                 ):
        super().__init__("tool_result", content, seconds_from_start)
        self.tool_call_id = tool_call_id
        self.type = result_type
        self.result = result

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "role": self.role,
            "content": self.content,
            "seconds_from_start": self.seconds_from_start,
            "tool_call_id": self.tool_call_id,
            "type": self.type,
            "result": self.result
        }



import re


def contains_full_sentence(text: str) -> bool:
    """
    Checks if there is at least one 'full sentence' in the string.
    A 'full sentence' must:
    1) Start with a capital letter,
    2) End with one of '.', '!', '?', '...',
    3) Followed by either the end of the text or by a space and a capital letter.
    """
    pattern = re.compile(
        r'[A-Z]'           # Start with uppercase letter
        r'[^.?!]*?'        # Then any characters, non-greedy
        r'(?:\.\.\.|[.?!])'  # End with '...', or '.', or '?', or '!'
        r'(?:\s+[A-Z]|$)'  # Followed by space + capital letter or end of string
    )
    return bool(pattern.search(text))


import os

from dotenv import load_dotenv

load_dotenv()

CONNEXITY_URL = os.environ.get('CONNEXITY_URL', default="https://connexity-gateway-owzhcfagkq-uc.a.run.app/process/sdk")
CONNEXITY_METRICS_URL = os.environ.get('CONNEXITY_METRICS_URL', default="https://connexity-gateway-owzhcfagkq-uc.a.run.app/process/sdk/llm_latency")


import asyncio
from typing import List, Optional, Literal

from connexity.calls.messages.interface_message import InterfaceMessage
from connexity.utils.send_data import send_data


class BaseCall:

    def __init__(self,
                 sid: str,
                 call_type: Literal["inbound", "outbound", "web"],
                 api_key: str,
                 agent_id: str,
                 user_phone_number: Optional[str] = None,
                 agent_phone_number: Optional[str] = None,
                 created_at=None,
                 voice_engine: Optional[str] = None,
                 transcriber: Optional[str] = None,
                 voice_provider: Optional[str] = None,
                 llm_model: Optional[str] = None,
                 llm_provider: Optional[str] = None,
                 phone_call_provider: Optional[str] = None,
                 stream: Optional[bool] = False
                 ):
        self._sid = sid
        self._call_type = call_type
        self._user_phone_number = user_phone_number
        self._agent_phone_number = agent_phone_number
        self._created_at = created_at
        self._agent_id = agent_id
        self._api_key = api_key
        self._voice_engine = voice_engine
        self._transcriber = transcriber
        self._voice_provider = voice_provider
        self._llm_model = llm_model
        self._llm_provider = llm_provider
        self._phone_call_provider = phone_call_provider
        self._stream = stream

        self._duration_in_seconds = None
        self._recording_url = None
        self._messages: List[InterfaceMessage] = []

    async def initialize(self):
        """Async method to handle post-init operations."""
        await self._send_data_to_connexity(update_type="status_update", status="in_progress", first=True)

    async def register_message(self, message: InterfaceMessage):
        self._messages.append(message)

        await self._send_data_to_connexity(update_type="conversation_update")

    async def update_last_message(self, message: InterfaceMessage):
        self._messages[-1] = message

        await self._send_data_to_connexity(update_type="conversation_update")
    
    async def init_post_call_data(self, recording_url, duration_in_seconds):
        self._recording_url = recording_url
        self._duration_in_seconds = duration_in_seconds

        await self._send_data_to_connexity(update_type="conversation_update")
        await self._send_data_to_connexity(update_type="status_update", status="completed")
        await self._send_data_to_connexity(update_type="end_of_call")

    def _to_dict(self) -> dict:
        """
        Returns a JSON representation of the BaseCall instance.

        Returns:
            str: JSON string of the instance.
        """
        return {
            'sid': self._sid,
            'call_type': self._call_type,
            'user_phone_number': self._user_phone_number,
            'agent_phone_number': self._agent_phone_number,
            'created_at': self._created_at.isoformat() if hasattr(self._created_at, 'isoformat') else self._created_at,
            'agent_id': self._agent_id,
            'voice_engine': self._voice_engine,
            'transcriber': self._transcriber,
            'voice_provider': self._voice_provider,
            'llm_model': self._llm_model,
            'llm_provider': self._llm_provider,
            'phone_call_provider': self._phone_call_provider,
            'duration_in_seconds': self._duration_in_seconds,
            'recording_url': self._recording_url,
            'messages': [message.to_dict() for message in self._messages] if self._messages else []
        }

    async def _send_data_to_connexity(self, update_type: str, status: Optional[str] = None, first=False):
        """
        Sends accumulated data to Connexity's backend.

        Returns:
            dict: Response from the Connexity API.
        """

        data = self._to_dict()
        data['event_type'] = update_type
        data["status"] = status

        if self._stream or self._recording_url or self._duration_in_seconds or first:
            await asyncio.create_task(send_data(data, self._api_key))



from connexity.CONST import CONNEXITY_URL

import aiohttp


async def send_data(answer_dict, api_key, url=CONNEXITY_URL):

    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers={"X-API-KEY": api_key}, json=answer_dict) as response:
            if response.status != 200:
                print(f"Coonexity SDK: Failed to send data: {response.status}")
                print(response, flush=True)
            else:
                print(f"Coonexity SDK: Data sent successfully: {response.status}", flush=True)
